SystemOrganization addCategory: #'Ma client server-Client API'!
SystemOrganization addCategory: #'Ma client server-Exceptions'!
SystemOrganization addCategory: #'Ma client server-Info'!
SystemOrganization addCategory: #'Ma client server-Server API'!

MaObject subclass: #MaClientServerPreferences
	instanceVariableNames: ''
	classVariableNames: 'Debug'
	poolDictionaries: ''
	category: 'Ma client server-Server API'!

!MaClientServerPreferences commentStamp: 'cmm 2/1/2005 15:27' prior: 0!
Control your Ma client server preferences with this class.!

!MaClientServerPreferences class methodsFor: 'accessing' stamp: 'cmm 1/2/2003 14:35'!
debug

	^Debug! !

!MaClientServerPreferences class methodsFor: 'accessing' stamp: 'cmm 1/2/2003 14:53'!
debug: aBoolean

	Debug _ aBoolean! !

!MaClientServerPreferences class methodsFor: 'private' stamp: 'cmm 1/2/2003 14:35'!
initialize

	super initialize.
	Debug _ false! !

MaObject subclass: #MaClientSocket
	instanceVariableNames: 'hostAddress port socket byteArray timeoutSeconds stat'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma client server-Client API'!

!MaClientSocket commentStamp: 'cmm 2/1/2005 15:29' prior: 0!
An easy-to-use class to send ByteArray or String requests to a socket listening on the network.  The other socket is listening using MaServerSocket.!

!MaClientSocket class methodsFor: 'creation' stamp: 'cmm 7/18/2005 13:19'!
hostAddress: aByteArray port: anInteger

	Socket initializeNetwork.
	^self new
		hostAddress: aByteArray ;
		port: anInteger ;
		yourself! !

!MaClientSocket class methodsFor: 'initializing' stamp: 'cmm 5/2/2003 15:08'!
initialize

	Smalltalk addToStartUpList: self! !

!MaClientSocket class methodsFor: 'initializing' stamp: 'cmm 5/2/2003 15:11'!
startUp: aBoolean

	super startUp: aBoolean.
	Socket initializeNetwork! !

!MaClientSocket methodsFor: 'private' stamp: 'cmm 6/3/2008 13:15'!
connect
	| tries startTryingToConnectSocket |
	socket 
		ifNotNil: 
			[ MagmaSoftwareError signal: 'client trying to make multiple connects?' ]
		ifNil: 
			[ socket := Socket newTCP.
			tries := 0.
			
			[ startTryingToConnectSocket := Time millisecondClockValue.
			socket 
				connectTo: hostAddress
				port: port.
			stat timeToConnectSocketHistory addValue: (Time millisecondsSince: startTryingToConnectSocket) ] 
				on: NetworkError
				do: 
					[ : error | 
					"I can't remember why I put in this retry logic, but I don't see any problem with keeping it."
					tries < 3
						ifTrue: 
							[ MaNetworkNotification signal: 'Socket couldn''t connect, retrying.'.
							(Delay forMilliseconds: 1000) wait.
							tries := tries + 1.
							error retry ]
						ifFalse: [ MaNetworkError signal: error messageText ] ].
			socket ]! !

!MaClientSocket methodsFor: 'private' stamp: 'cmm 5/29/2008 09:53'!
disconnect
	socket closeAndDestroy: 5.
	socket := nil! !

!MaClientSocket methodsFor: 'private' stamp: 'cmm 11/26/2004 09:45'!
ensureByteArrayCanHold: anInteger

	anInteger > 10000000 ifTrue: [ MaSecurityViolation signal: 'Possible attack.' ].
	anInteger > byteArray size
		ifTrue:
			[ | newByteArray |
			newByteArray _ ByteArray new: anInteger.
			newByteArray
				replaceFrom: 1
				to: byteArray size
				with: byteArray
				startingAt: 1.
			byteArray _ newByteArray ]! !

!MaClientSocket methodsFor: 'private' stamp: 'cmm 6/2/2008 15:53'!
ensureConnected
	"Ensure the primitive Socket is connected."
	socket 
		ifNil: [ self connect ]
		ifNotNil: 
			[ socket isConnected ifFalse: 
				[ self
					disconnect ;
					connect ] ]! !

!MaClientSocket methodsFor: 'private' stamp: 'cmm 9/20/2002 14:28'!
headerSize

	^4! !

!MaClientSocket methodsFor: 'accessing' stamp: 'cmm 9/20/2002 12:24'!
hostAddress

	^hostAddress! !

!MaClientSocket methodsFor: 'accessing' stamp: 'cmm 9/20/2002 12:23'!
hostAddress: aByteArray

	hostAddress _ aByteArray! !

!MaClientSocket methodsFor: 'private' stamp: 'cmm 10/1/2002 19:43'!
hostAndPortString

	^'host: ', hostAddress maAsIpString, ' port: ', port printString! !

!MaClientSocket methodsFor: 'initializing' stamp: 'cmm 5/7/2008 16:09'!
initialize
	super initialize.
	byteArray := ByteArray new: 1000.
	timeoutSeconds := 30.
	stat := MaClientStat new! !

!MaClientSocket methodsFor: 'private' stamp: 'cmm 6/8/2008 23:05'!
loadByteArrayWith: aByteArray startingAt: indexPosition count: anInteger 
	| sendSize |
	sendSize := self headerSize + aByteArray size.
	self ensureByteArrayCanHold: sendSize.
	byteArray 
		replaceFrom: self headerSize + 1
		to: sendSize
		with: aByteArray
		startingAt: indexPosition.
	byteArray 
		maUint: 32
		at: 0
		put: sendSize! !

!MaClientSocket methodsFor: 'accessing' stamp: 'cmm 9/20/2002 12:23'!
port

	^port! !

!MaClientSocket methodsFor: 'accessing' stamp: 'cmm 9/20/2002 12:23'!
port: anInteger

	port _ anInteger! !

!MaClientSocket methodsFor: 'private' stamp: 'cmm 6/8/2008 23:14'!
receiveInto: aByteArray 
	"First, we receive into *my* byteArray, then I paste it into aByteArray."
	| bytesExpected bytesReceived answer startedWaitingForResponse startedReceivingResponse timeToReceiveResponse |
	bytesReceived := 0.
	bytesExpected := nil.
	"We track two separate timing stats from this method, first the #timeWaitingForResponsesHistory is the time to receive the *initial bytes* of the response.  Second, the #timeReceivingResponsesHistory measures the time to actually *download* the response bytes, calculated by subtracting the value last recorded in the timeWaitingForResponsesHistory."
	startedReceivingResponse := Time millisecondClockValue.
	[ bytesExpected isNil or: [ bytesReceived < bytesExpected ] ] whileTrue: 
		[ startedWaitingForResponse ifNil: [ startedWaitingForResponse := Time millisecondClockValue ].
		socket waitForDataFor: timeoutSeconds.
		startedWaitingForResponse isInteger ifTrue: 
			[ stat timeWaitingForResponsesHistory addValue: (Time millisecondsSince: startedWaitingForResponse).
			startedWaitingForResponse := 'initial wait captured.'	"Deactivate it now by setting it to this string." ].
		
		[ bytesReceived := bytesReceived + (socket 
				receiveSomeDataInto: byteArray
				startingAt: bytesReceived + 1) ] 
			on: Error
			do: 
				[ : err | 
				err messageText = 'a primitive has failed' ifTrue: 
					[ err resignalAs: (MaNetworkError new
							messageText: 'Network failure while receiving response.' ;
							yourself) ] ].
		bytesExpected ifNil: 
			[ bytesReceived >= self headerSize ifTrue: 
				[ bytesExpected := byteArray 
					maUint: 32
					at: 0.
				self ensureByteArrayCanHold: bytesExpected ] ] ].
	"I think it's safe to assume here that the stats will be active."
	timeToReceiveResponse := (Time millisecondsSince: startedReceivingResponse) - stat timeWaitingForResponsesHistory last.
	stat timeReceivingResponsesHistory addValue: timeToReceiveResponse.
	"We should now have the full buffer in my byteArray, paste it into the answer ByteArray."
	stat timeToLoadBuffersHistory addValue: (Time millisecondsToRun: 
			[ answer := bytesReceived - self headerSize > aByteArray size 
				ifTrue: [ ByteArray new: bytesReceived - self headerSize ]
				ifFalse: [ aByteArray ].
			(answer 
				replaceFrom: 1
				to: bytesReceived - self headerSize
				with: byteArray
				startingAt: self headerSize + 1) ]).
	^ answer lzUncompressed! !

!MaClientSocket methodsFor: 'private' stamp: 'cmm 6/2/2008 22:51'!
send: anInteger 
	| startedSending |
	self ensureConnected.
	startedSending := Time millisecondClockValue.
	socket 
		sendData: byteArray
		count: anInteger.
	socket waitForSendDoneFor: timeoutSeconds.
	socket sendDone ifFalse: 
		[ MaNetworkError signal: 'Connection to ' , self hostAndPortString , ' lost.' ].
	"Only include successful sends to measure real network performance, not timeouts perhaps due to user-error."
	stat timeToSendRequestsHistory addValue: (Time millisecondsSince: startedSending)! !

!MaClientSocket methodsFor: 'send' stamp: 'cmm 1/30/2003 22:08'!
sendData: requestByteArray

	| answer |
	answer _ ByteArray new: 500.
	^self
		sendData: requestByteArray
		startingAt: 1
		count: requestByteArray size
		waitForReplyIn: answer! !

!MaClientSocket methodsFor: 'send' stamp: 'cmm 6/8/2008 22:35'!
sendData: requestByteArray startingAt: startIndex count: numberOfBytes waitForReplyIn: answerByteArray 
	"Perform a synchronous TCP call, sending count bytes of requestByteArray from position startIndex to my peer socket and then wait up to my timeout period for a reply, which I'll put in answerByteArray.  If I run out of room in answerByteArray, I'll create you a new one and return it, otherwise you'll get your own answerByteArray back.  If communications are disrupted, I'll signal a MaNetworkError."
	| tries compressedBa|
	stat timeToLoadBuffersHistory addValue: 
		(Time millisecondsToRun: 
			[ self 
				loadByteArrayWith: (compressedBa := requestByteArray lzCompressed)
				startingAt: startIndex
				count: numberOfBytes ]).
	"Do a retry logic to help with the case where the server has dropped only momentarily.  Sockets will be invalid but, unfortunately, they don't always correctly report they are invalid, as if they don't 'realize' it yet."
	tries := 0.
	[ self send: compressedBa size + self headerSize ] 
		on: Error
		do: 
			[ : sendErr | 
			(sendErr messageText = 'a primitive has failed' or: [ sendErr isKindOf: NetworkError ]) 
				ifTrue: 
					[ tries < 3 
						ifTrue: 
							[ tries := tries + 1.
							sendErr retry ]
						ifFalse: 
							[ sendErr resignalAs: (MaNetworkError new messageText: sendErr messageText , ' while sending request.') ] ]
				ifFalse: 
					[ sendErr resignalAs: (MaNetworkError new messageText: sendErr messageText , ' while sending request.') ] ].
	"Retry logic is not appropriate on the receive side."
	^ [ self receiveInto: answerByteArray ] 
		on: NetworkError
		do: 
			[ : receiveErr | 
			receiveErr resignalAs: (MaNetworkError new messageText: receiveErr messageText , ' while receiving response.') ]! !

!MaClientSocket methodsFor: 'private' stamp: 'cmm 5/6/2008 23:17'!
stat
	^stat! !

!MaClientSocket methodsFor: 'private' stamp: 'cmm 5/6/2008 23:18'!
stat: aMaClientStat
	stat := aMaClientStat! !

!MaClientSocket methodsFor: 'accessing' stamp: 'cmm 9/20/2002 15:10'!
timeoutSeconds: anInteger

	timeoutSeconds _ anInteger! !

MaObject subclass: #MaServer
	instanceVariableNames: 'console'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma client server-Server API'!

!MaServer commentStamp: 'cmm 2/1/2005 15:35' prior: 0!
Abstract superclass for the "server engine".  I know a console which I send output to.!

MaServer subclass: #MaRequestServer
	instanceVariableNames: 'processor'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma client server-Server API'!

!MaRequestServer commentStamp: 'cmm 2/20/2005 17:15' prior: 0!
Abstract class whose subclass instances handle the server portion of the client-server activity.  The processor should be an object that responds to #value:.  The argument is the actual request object and it should answer the result object.  Requests and responses can be arbitrary objects, not just byte-objects like Strings and ByteArrays, although you must set up your #protocol.!

MaRequestServer subclass: #MaLocalRequestServer
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma client server-Server API'!

!MaLocalRequestServer commentStamp: 'cmm 5/8/2008 21:21' prior: 0!
Use this server as a stub when you are running your object-request server locally within the same image.!

!MaLocalRequestServer class methodsFor: 'creation' stamp: 'cmm 9/19/2004 22:38'!
processor: oneArgBlockOrProcessor

	^ self new 
		processor: oneArgBlockOrProcessor ;
		yourself! !

!MaLocalRequestServer methodsFor: 'as yet unclassified' stamp: 'cmm 5/2/2005 22:25'!
isRunning

	^ true! !

!MaRequestServer methodsFor: 'private' stamp: 'cmm 5/1/2008 20:55'!
connectionRequest

	^'connect'! !

!MaRequestServer methodsFor: 'initializing' stamp: 'cmm 1/5/2003 23:12'!
initialize

	super initialize.
	console _ MaRequestServerConsole new! !

!MaRequestServer methodsFor: 'testing' stamp: 'cmm 2/20/2005 17:18'!
isRunning

	self subclassResponsibility! !

!MaRequestServer methodsFor: 'private' stamp: 'cmm 5/1/2008 18:30'!
processRequest: aMaServerRequest 
	"Here's your request object, do something with it and answer any object, though for remote connections only, every class throughout the entire request and responses graph must be part of your #protocol."
	| answer |
	console noteAboutToProcessRequest: aMaServerRequest.
	answer := 
	[ aMaServerRequest = self connectionRequest 
		ifTrue: [ self connectionInfo ]
		ifFalse: [ processor value: aMaServerRequest ] ] 
		on: MaNotification
		do: 
			[ : noti | 
			console note: noti messageText.
			noti resume ].
	console noteRequestProcessed: aMaServerRequest.
	^ answer! !

!MaRequestServer methodsFor: 'accessing' stamp: 'cmm 12/8/2004 13:34'!
processor
	^ processor! !

!MaRequestServer methodsFor: 'private' stamp: 'cmm 9/19/2004 22:10'!
processor: oneArgBlock
	"oneArgBlock takes the request object, answers the response object you want to send back to the client.  For remote, these can only be objects defined in the protocol."

	processor _ oneArgBlock! !

MaRequestServer subclass: #MaTcpRequestServer
	instanceVariableNames: 'socket serializer nextId protocolByteArray'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma client server-Server API'!

!MaTcpRequestServer commentStamp: 'cmm 2/1/2005 15:40' prior: 0!
The server to use when you want to allow requests and responses to be any object and run on a remote server.!

!MaTcpRequestServer class methodsFor: 'creation' stamp: 'cmm 10/8/2002 22:50'!
protocol: collectionOfClasses
	"collectionOfClasses are all request classes and all response classes any client can expect to encounter in the full graph of the requests and responses"

	^self new protocol: collectionOfClasses! !

!MaTcpRequestServer methodsFor: 'private' stamp: 'cmm 10/11/2002 00:24'!
connectionInfo

	^Array
		with: self getNextId
		with: self protocolByteArray! !

!MaTcpRequestServer methodsFor: 'private' stamp: 'cmm 10/11/2002 00:25'!
getNextId

	^nextId _ nextId + 1! !

!MaTcpRequestServer methodsFor: 'private' stamp: 'cmm 11/26/2004 13:15'!
initialize

	super initialize.
	socket _
		MaServerSocket new
			server: self;
			yourself.
	nextId _ 1.
	self initializeSerializer! !

!MaTcpRequestServer methodsFor: 'private' stamp: 'cmm 6/10/2005 12:21'!
initializeSerializer

	serializer _ MaObjectSerializer new! !

!MaTcpRequestServer methodsFor: 'testing' stamp: 'cmm 2/20/2005 17:18'!
isRunning

	^ socket isRunning! !

!MaTcpRequestServer methodsFor: 'private' stamp: 'cmm 5/1/2008 18:16'!
materializeRequest: requestByteArray 
	| timeToMaterialize requestObject |
	timeToMaterialize := Time millisecondsToRun: 
		[ serializer resetOidManager.
		requestObject := serializer materializeGraph: requestByteArray ].
	self noteTimeToMaterializeRequest: timeToMaterialize.
	^ requestObject! !

!MaTcpRequestServer methodsFor: 'start / stop' stamp: 'cmm 4/29/2003 23:28'!
processOn: portNumber
	"Listen on portNumber in the background.  For each client request received, evaluate my processBlock.  I queue all requests so that your block will never be evaluating more than one client-request at a time."

	socket server: self.
	socket listenOn: portNumber! !

!MaTcpRequestServer methodsFor: 'start / stop' stamp: 'cmm 2/20/2005 18:47'!
processOn: portNumber using: yourProcessor
	"Listen on portNumber in the background.  For each client request received, value: yourProcessor (which must respond to #value:).  I queue all requests and ensure your yourProcessor will never be evaluating more than one client-request at a time."

	processor _ yourProcessor.
	self processOn: portNumber! !

!MaTcpRequestServer methodsFor: 'private' stamp: 'cmm 5/1/2008 18:28'!
processRequestByteArray: requestByteArray 
	"Materialize the object serialized in requestByteArray, and serialize the response I get from telling my processor to process it."
	| requestObject responseObject |
	[ requestObject := self materializeRequest: requestByteArray ] 
		on: Error
		do: 
			[ : exc | 
			MaClientServerPreferences debug 
				ifTrue: [ exc pass ]
				ifFalse: 
					[ ^ self serializeResponse: (MaMalformedRequestError new
							messageText: exc messageText ;
							yourself) ] ].
	responseObject := self processRequest: requestObject.
	^ self serializeResponse: responseObject! !

!MaTcpRequestServer methodsFor: 'private' stamp: 'cmm 8/23/2007 16:26'!
protocol: aCollection
	"Establish the class ids of classes in the protocol."
	aCollection do: [ : each | serializer classIdManager addNewClass: each ].
	"Protocol established, now disallow any further new classes!!"
	serializer classIdManager beImmutable ! !

!MaTcpRequestServer methodsFor: 'private' stamp: 'cmm 5/1/2008 18:24'!
protocolByteArray
	^ protocolByteArray ifNil: [ protocolByteArray := serializer classDefinitionsByteArray ]! !

!MaTcpRequestServer methodsFor: 'private' stamp: 'cmm 5/1/2008 19:00'!
serializeResponse: anObject 
	| timeToSerializeResponse response |
	timeToSerializeResponse := Time millisecondsToRun: 
		[ serializer resetOidManager.
		response := (serializer serializeGraph: anObject) trimmedByteArray ].
	console noteTimeToSerializeResponse: timeToSerializeResponse.
	^ response! !

!MaTcpRequestServer methodsFor: 'start / stop' stamp: 'cmm 12/31/2002 14:01'!
shutdown

	socket shutdown.
	"socket shutdown immediately disallows any more requests, but blocks until pending requests get processed.  Those remaining requests which are not finished writing responses can and will continue to write, but the server is no longer needed and able to shutdown."
	console noteServerShutdown! !

!MaTcpRequestServer methodsFor: 'accessing' stamp: 'cmm 10/8/2002 22:23'!
timeoutSeconds

	^socket timeoutSeconds! !

!MaTcpRequestServer methodsFor: 'accessing' stamp: 'cmm 10/8/2002 22:23'!
timeoutSeconds: anInteger

	socket timeoutSeconds: anInteger! !

!MaServer methodsFor: 'console' stamp: 'cmm 10/7/2002 23:45'!
console: aMaServerConsole

	console _ aMaServerConsole! !

!MaServer methodsFor: 'private' stamp: 'cmm 10/8/2002 23:05'!
initialize

	super initialize.
	console _ MaServerConsole new! !

!MaServer methodsFor: 'notifications' stamp: 'cmm 10/10/2002 00:29'!
note: aString
	"The socket is notifying us of a message."

	console note: aString! !

!MaServer methodsFor: 'notifications' stamp: 'cmm 5/1/2008 20:47'!
noteClientTimedOut: aMaRequestLifecycle 
	"Network issues, client socket is no longer communicating."
	console noteClientTimedOut: aMaRequestLifecycle! !

!MaServer methodsFor: 'notifications' stamp: 'cmm 5/1/2008 20:53'!
noteException: anException whileReceivingRequestFrom: aMaRequestLifecycle 
	"anException was signaled when trying to read from aSocket.  Maybe they disconnected?"
	console 
		noteException: anException
		whileReceivingRequestFrom: aMaRequestLifecycle! !

!MaServer methodsFor: 'notifications' stamp: 'cmm 10/10/2002 00:30'!
noteNoLongerAcceptingRequests
	"Called when the socket has previously been told to shutdown and can no longer accept requests."

	console noteNoLongerAcceptingRequests! !

!MaServer methodsFor: 'notifications' stamp: 'cmm 10/10/2002 00:30'!
noteNoLongerExtractingRequests
	"Called when the socket has previously been told to shutdown and extracted the last pending sockets request."

	console noteNoLongerExtractingRequests! !

!MaServer methodsFor: 'notifications' stamp: 'cmm 10/10/2002 00:30'!
noteNoLongerProcessingRequests
	"Called when the socket has previously been told to shutdown and has processed the last pending request."

	console noteNoLongerProcessingRequests! !

!MaServer methodsFor: 'notifications' stamp: 'cmm 6/8/2008 11:29'!
noteNumberOfConnections: anInteger 
	"Report the number of known MaClientConnections."
	console noteNumberOfConnections: anInteger! !

!MaServer methodsFor: 'notifications' stamp: 'cmm 5/1/2008 17:54'!
noteRequestDequeued: aMaRequestLifecycle 
	"The response has just been popped off the requestQueue and about to be processed.  Consoles will want to note the time spent waiting in the requestQueue."
	console noteRequestDequeued: aMaRequestLifecycle! !

!MaServer methodsFor: 'notifications' stamp: 'cmm 5/1/2008 17:51'!
noteRequestQueued: aMaRequestLifecycle 
	"The aMaRequestLifecycle's now has a requestByteArray that has been extracted from its socket.  The requestByteArray has now been appended to the end of the requestQueue."
	console noteRequestQueued: aMaRequestLifecycle! !

!MaServer methodsFor: 'notifications' stamp: 'cmm 5/1/2008 19:31'!
noteResponseSent: aMaRequestLifecycle 
	"Successfully sent the aByteArray to aSocket."
	console noteResponseSent: aMaRequestLifecycle! !

!MaServer methodsFor: 'notifications' stamp: 'cmm 10/10/2002 00:32'!
noteSendResponse: aByteArray timedOutOn: aSocket
	"aSocket must have disconnected or something, we couldn't get the response to them."

	console
		noteSendResponse: aByteArray
		timedOutOn: aSocket! !

!MaServer methodsFor: 'notifications' stamp: 'cmm 10/10/2002 00:31'!
noteSendingResponse: aByteArray to: aSocket
	"a ByteArray is about to be sent to aSocket."

	console
		noteSendingResponse: aByteArray 
		to: aSocket! !

!MaServer methodsFor: 'notifications' stamp: 'cmm 5/1/2008 17:34'!
noteSocketDequeued: aMaRequestLifecycle 
	"aMaRequestLifecycle's socket has been accepted by the server, and a process forked to extracted its byteArray request."
	console noteSocketDequeued: aMaRequestLifecycle! !

!MaServer methodsFor: 'notifications' stamp: 'cmm 10/10/2002 00:32'!
noteSocketQueued: clientSocket
	"clientSocket has been accepted by the server, and a process forked to extracted its byteArray request."

	console noteSocketQueued: clientSocket! !

!MaServer methodsFor: 'notifications' stamp: 'cmm 5/1/2008 16:44'!
noteTimeToExtractRequestBytes: milliseconds 
	"milliseconds is the amount of time required to extract all of the data out of the socket."
	console noteTimeToExtractRequestBytes: milliseconds! !

!MaServer methodsFor: 'notifications' stamp: 'cmm 6/8/2008 21:40'!
noteTimeToLoadBuffers: milliseconds 
	"The number of milliseconds spent compressing and uncompressing ByteArrays for more efficient network transport."
	console noteTimeToLoadBuffers: milliseconds! !

!MaServer methodsFor: 'notifications' stamp: 'cmm 5/1/2008 18:14'!
noteTimeToMaterializeRequest: milliseconds 
	"milliseconds is the amount of time required to extract all of the data out of the socket."
	console noteTimeToMaterializeRequest: milliseconds! !

!MaServer methodsFor: 'private' stamp: 'cmm 8/23/2005 20:08'!
processRequestByteArray: aByteArray
	"Default answer empty string."

	^ String maEmpty! !

MaObject subclass: #MaServerConsole
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma client server-Server API'!

!MaServerConsole commentStamp: 'cmm 2/1/2005 15:34' prior: 0!
I am used to receive messages from the server.!

MaServerConsole subclass: #MaRequestServerConsole
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma client server-Server API'!

!MaRequestServerConsole commentStamp: 'cmm 2/1/2005 15:34' prior: 0!
I am used to receive messages from the server.!

!MaRequestServerConsole methodsFor: 'notifications' stamp: 'cmm 5/1/2008 15:40'!
noteAboutToProcessRequest: yourRequestObject 
	"optional subclass responsibility"
	! !

!MaRequestServerConsole methodsFor: 'notifications' stamp: 'cmm 1/5/2003 22:41'!
noteRequestProcessed: yourRequestObject
	"optional subclass responsibility"! !

MaRequestServerConsole subclass: #MaTranscriptConsole
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma client server-Server API'!

!MaTranscriptConsole commentStamp: 'cmm 2/1/2005 15:41' prior: 0!
A default console that prints messages to the Transcript.!

!MaTranscriptConsole methodsFor: 'output' stamp: 'cmm 10/1/2002 22:49'!
newLine

	Transcript cr! !

!MaTranscriptConsole methodsFor: 'notifications' stamp: 'cmm 10/1/2002 23:33'!
note: aString

	self
		newLine;
		showMessage: aString! !

!MaTranscriptConsole methodsFor: 'notifications' stamp: 'cmm 5/1/2008 20:53'!
noteException: anException onSocket: aMaRequestLifecycle 
	"anException was signaled when trying to read from aSocket.  Maybe they disconnected?"
	self
		newLine ;
		showSocket: aMaRequestLifecycle socket ;
		showSeparator ;
		showMessage: (anException messageText isEmpty 
				ifTrue: [ anException species name ]
				ifFalse: [ anException messageText ])! !

!MaTranscriptConsole methodsFor: 'notifications' stamp: 'cmm 5/1/2008 20:53'!
noteException: anException whileReceivingRequestFrom: aMaRequestLifecycle 
	"anException was signaled when trying to read from aSocket.  Maybe they disconnected?"
	self 
		noteException: anException
		onSocket: aMaRequestLifecycle! !

!MaTranscriptConsole methodsFor: 'notifications' stamp: 'cmm 10/9/2002 23:07'!
noteNoLongerProcessingRequests
	"Called when the socket has previously been told to shutdown and has processed the last pending request."

	self newLine; showMessage: 'I shall process no more requests.'! !

!MaTranscriptConsole methodsFor: 'notifications' stamp: 'cmm 5/1/2008 19:33'!
noteResponseSent: aMaRequestLifecycle 
	"Successfully sent the aMaRequestLifecycle's responseByteArray to its socket."
	self
		newLine ;
		showMessage: 'sent ' , aMaRequestLifecycle responseByteArray size printString , ' bytes to ' , aMaRequestLifecycle socket remoteAddress maAsIpString! !

!MaTranscriptConsole methodsFor: 'notifications' stamp: 'cmm 10/1/2002 23:15'!
noteSendResponse: aByteArray timedOutOn: aSocket
	"aSocket must have disconnected or something, we couldn't get the response to them."

	self
		newLine;
		showMessage: 'timed out trying to send ', aByteArray size printString, ' bytes to ', aSocket remoteAddress maAsIpString! !

!MaTranscriptConsole methodsFor: 'notifications' stamp: 'cmm 10/1/2002 23:15'!
noteSendingResponse: aByteArray to: aSocket
	"a ByteArray is about to be sent to aSocket."

	self
		newLine;
		showMessage: 'about to send ', aByteArray size printString, ' bytes to ', aSocket remoteAddress maAsIpString! !

!MaTranscriptConsole methodsFor: 'notifications' stamp: 'cmm 10/1/2002 23:16'!
noteSocketQueued: clientSocket
	"clientSocket has been accepted by the server, and a process forked to extracted its byteArray request."

	self
		newLine;
		showMessage: 'queued ', clientSocket remoteAddress maAsIpString ! !

!MaTranscriptConsole methodsFor: 'output' stamp: 'cmm 10/1/2002 22:49'!
showMessage: aString

	Transcript show: aString! !

!MaTranscriptConsole methodsFor: 'output' stamp: 'cmm 10/1/2002 22:53'!
showSeparator

	Transcript show: ' - '! !

!MaTranscriptConsole methodsFor: 'output' stamp: 'cmm 10/1/2002 22:51'!
showSocket: aSocket

	Transcript show: aSocket remoteAddress maAsIpString! !

!MaServerConsole methodsFor: 'notifications' stamp: 'cmm 10/7/2002 23:51'!
note: aString
	"The socket is notifying us of a message."! !

!MaServerConsole methodsFor: 'notifications' stamp: 'cmm 5/1/2008 20:44'!
noteClientTimedOut: aMaRequestLifecycle 
	"Network issues , client socket is no longer communicating"! !

!MaServerConsole methodsFor: 'notifications' stamp: 'cmm 6/5/2008 21:27'!
noteException: anException whileReceivingRequestFrom: aMaRequestLifecycle
	"anException was signaled when trying to read from aSocket."! !

!MaServerConsole methodsFor: 'notifications' stamp: 'cmm 10/7/2002 23:51'!
noteNoLongerAcceptingRequests
	"Called when the socket has previously been told to shutdown and can no longer accept requests."! !

!MaServerConsole methodsFor: 'notifications' stamp: 'cmm 10/7/2002 23:51'!
noteNoLongerExtractingRequests
	"Called when the socket has previously been told to shutdown and extracted the last pending sockets request."! !

!MaServerConsole methodsFor: 'notifications' stamp: 'cmm 10/7/2002 23:51'!
noteNoLongerProcessingRequests
	"Called when the socket has previously been told to shutdown and has processed the last pending request."! !

!MaServerConsole methodsFor: 'notifications' stamp: 'cmm 6/8/2008 11:30'!
noteNumberOfConnections: anInteger 
	"Report the number of known MaClientConnections."! !

!MaServerConsole methodsFor: 'notifications' stamp: 'cmm 5/1/2008 17:54'!
noteRequestDequeued: aMaRequestLifecycle 
	! !

!MaServerConsole methodsFor: 'notifications' stamp: 'cmm 5/1/2008 17:51'!
noteRequestQueued: aMaRequestLifecycle
! !

!MaServerConsole methodsFor: 'notifications' stamp: 'cmm 5/1/2008 19:32'!
noteResponseSent: aMaRequestLifecycle
	"Successfully sent the aMaRequestLifecycle's responseByteArray to its socket."! !

!MaServerConsole methodsFor: 'notifications' stamp: 'cmm 10/7/2002 23:51'!
noteSendResponse: aByteArray timedOutOn: aSocket
	"aSocket must have disconnected or something, we couldn't get the response to them."! !

!MaServerConsole methodsFor: 'notifications' stamp: 'cmm 10/7/2002 23:51'!
noteSendingResponse: aByteArray to: aSocket
	"a ByteArray is about to be sent to aSocket."! !

!MaServerConsole methodsFor: 'notifications' stamp: 'cmm 10/8/2002 23:10'!
noteServerShutdown! !

!MaServerConsole methodsFor: 'notifications' stamp: 'cmm 1/5/2003 18:18'!
noteSocketDequeued: clientSocket
	"clientSocket is about processed by the server."! !

!MaServerConsole methodsFor: 'notifications' stamp: 'cmm 10/7/2002 23:51'!
noteSocketQueued: clientSocket
	"clientSocket has been accepted by the server, and a process forked to extracted its byteArray request."
! !

!MaServerConsole methodsFor: 'notifications' stamp: 'cmm 5/1/2008 16:44'!
noteTimeToExtractRequestBytes: milliseconds
	"Time to extract the request bytes out of the primitive TCP/IP socket."! !

!MaServerConsole methodsFor: 'notifications' stamp: 'cmm 6/8/2008 21:39'!
noteTimeToLoadBuffers: milliseconds
	"The number of milliseconds spent compressing and uncompressing ByteArrays for more efficient network transport."
! !

!MaServerConsole methodsFor: 'notifications' stamp: 'cmm 5/1/2008 18:12'!
noteTimeToMaterializeRequest: milliseconds
	"The time to materialize a serialized request object."! !

!MaServerConsole methodsFor: 'notifications' stamp: 'cmm 5/1/2008 18:55'!
noteTimeToSerializeResponse: milliseconds
	"The time to materialize a serialized request object."! !

MaObject subclass: #MaServerLink
	instanceVariableNames: 'stat'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma client server-Client API'!

!MaServerLink commentStamp: 'cmm 2/1/2005 15:36' prior: 0!
Abstract superclass representing a clients link to a server.  A client uses this link to submit request objects via #submit:.!

MaServerLink subclass: #MaLocalServerLink
	instanceVariableNames: 'localRequestServer'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma client server-Client API'!

!MaLocalServerLink commentStamp: 'cmm 2/1/2005 15:32' prior: 0!
This is your clients interface to the server when the server is running locally (e.g., with MaLocalRequestServer).!

!MaLocalServerLink class methodsFor: 'as yet unclassified' stamp: 'cmm 3/30/2002 22:50'!
localRequestServer: aMaLocalRequestServer

	^self new localRequestServer: aMaLocalRequestServer
! !

!MaLocalServerLink methodsFor: 'testing' stamp: 'cmm 10/8/2002 23:59'!
isLocal

	^true! !

!MaLocalServerLink methodsFor: 'accessing' stamp: 'cmm 1/14/2005 15:11'!
linkInformation

	^ 'local-link to ', localRequestServer processor pathName! !

!MaLocalServerLink methodsFor: 'private' stamp: 'cmm 7/2/2004 13:17'!
localRequestServer

	^ localRequestServer! !

!MaLocalServerLink methodsFor: 'private' stamp: 'cmm 3/30/2002 22:52'!
localRequestServer: aMaLocalRequestServer

	localRequestServer _ aMaLocalRequestServer
! !

!MaLocalServerLink methodsFor: 'actions' stamp: 'cmm 3/30/2002 22:03'!
submit: aMaServerRequest

	^localRequestServer processRequest: aMaServerRequest
! !

!MaServerLink methodsFor: 'actions' stamp: 'cmm 6/4/2008 10:06'!
connect
	"Remote links have connection code they need to run.."
	stat transition! !

!MaServerLink methodsFor: 'actions' stamp: 'cmm 5/7/2008 16:49'!
disconnect
	"Remote links need to clean up resources."
	stat release! !

!MaServerLink methodsFor: 'access' stamp: 'cmm 8/12/2007 20:36'!
hostAddress
	MaUserError signal: 'not remotly connected.'! !

!MaServerLink methodsFor: 'initialize-release' stamp: 'cmm 5/7/2008 16:10'!
initialize
	super initialize.
	stat := MaClientStat new! !

!MaServerLink methodsFor: 'testing' stamp: 'cmm 10/8/2002 23:59'!
isLocal

	^false! !

!MaServerLink methodsFor: 'override' stamp: 'cmm 7/2/2004 17:16'!
linkInformation

	self subclassResponsibility! !

!MaServerLink methodsFor: 'access' stamp: 'cmm 8/12/2007 20:38'!
port
	MaUserError signal: 'not remotly connected.'! !

!MaServerLink methodsFor: 'access' stamp: 'cmm 5/6/2008 23:14'!
stat: aMaClientStat
	stat := aMaClientStat! !

!MaServerLink methodsFor: 'actions' stamp: 'cmm 3/30/2002 22:01'!
submit: aMaServerRequest

	self subclassResponsibility! !

!MaServerLink methodsFor: 'access' stamp: 'cmm 10/16/2002 22:05'!
timeoutSeconds: anInteger
	"for interface compatibility with remote server links."! !

MaServerLink subclass: #MaTcpRequestServerLink
	instanceVariableNames: 'serializer socket submitGuard'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma client server-Client API'!

!MaTcpRequestServerLink commentStamp: 'cmm 2/1/2005 15:40' prior: 0!
This is your clients interface to the server when the server is running remotely (e.g., with MaTcpRequestServer).!

!MaTcpRequestServerLink class methodsFor: 'creation' stamp: 'cmm 10/8/2002 22:17'!
hostAddress: aByteArray port: anInteger
	"bufferSize is the size of the byteArray used to send requests and materialize responses."

	^self new
		socket:
			(MaClientSocket
				hostAddress: aByteArray
				port: anInteger);
		yourself! !

!MaTcpRequestServerLink methodsFor: 'connections' stamp: 'cmm 6/4/2008 10:07'!
connect
	| idAndProtocol |
	super connect.
	idAndProtocol _ self primSubmit: self connectionRequest.
	serializer classDefinitionsByteArray: (idAndProtocol at: 2).
	serializer classIdManager beImmutable! !

!MaTcpRequestServerLink methodsFor: 'private' stamp: 'cmm 10/16/2002 14:26'!
connectionRequest

	^'connect'! !

!MaTcpRequestServerLink methodsFor: 'connections' stamp: 'cmm 5/7/2008 16:49'!
disconnect
	super disconnect.
	socket disconnect! !

!MaTcpRequestServerLink methodsFor: 'private' stamp: 'cmm 5/7/2008 16:10'!
getByteArrayResponseFor: aMaServerRequest 
	| graphBuffer timeToSerialize |
	timeToSerialize := Time millisecondsToRun: 
		[ graphBuffer := serializer serializeGraph: aMaServerRequest ].
	stat timeToSerializeHistory addValue: timeToSerialize.
	^ socket 
		sendData: graphBuffer byteArray
		startingAt: 1
		count: graphBuffer bufferSize
		waitForReplyIn: serializer byteArray! !

!MaTcpRequestServerLink methodsFor: 'accessing' stamp: 'cmm 5/8/2002 23:24'!
hostAddress

	^socket hostAddress! !

!MaTcpRequestServerLink methodsFor: 'private' stamp: 'cmm 2/25/2008 20:29'!
initialize
	super initialize.
	self initializeSerializer.
	submitGuard _ Monitor new! !

!MaTcpRequestServerLink methodsFor: 'private' stamp: 'cmm 6/10/2005 12:21'!
initializeSerializer

	serializer _ MaObjectSerializer new! !

!MaTcpRequestServerLink methodsFor: 'connections' stamp: 'cmm 2/25/2008 22:02'!
isConnected
	^ serializer classIdManager isMutable not! !

!MaTcpRequestServerLink methodsFor: 'accessing' stamp: 'cmm 7/2/2004 17:16'!
linkInformation

	^ (self hostAddress->self port) printString! !

!MaTcpRequestServerLink methodsFor: 'accessing' stamp: 'cmm 5/8/2002 23:24'!
port

	^socket port! !

!MaTcpRequestServerLink methodsFor: 'private' stamp: 'cmm 5/9/2008 17:37'!
primSubmit: aMaServerRequest 
	^ submitGuard critical: 
		[ | answer responseByteArray beganSubmitting |
		beganSubmitting := Time millisecondClockValue.
		serializer resetOidManager.
		socket stat: stat.
		responseByteArray := self getByteArrayResponseFor: aMaServerRequest.
		serializer resetOidManager.
		stat timeToMaterializeHistory addValue: (Time millisecondsToRun: [ answer := serializer materializeGraph: responseByteArray ]).
		(stat timeToSubmitHistoryFor: aMaServerRequest class) addValue: (Time millisecondsSince: beganSubmitting).
		answer ]! !

!MaTcpRequestServerLink methodsFor: 'private' stamp: 'cmm 9/22/2002 21:32'!
socket: aMaClientSocket

	socket _ aMaClientSocket! !

!MaTcpRequestServerLink methodsFor: 'accessing' stamp: 'cmm 5/7/2008 16:06'!
stat
	^ stat! !

!MaTcpRequestServerLink methodsFor: 'actions' stamp: 'cmm 2/25/2008 21:57'!
submit: aMaServerRequest 
	^ submitGuard critical: 
		[ "ensure connected"
		self isConnected ifFalse: [ self connect ].
		self primSubmit: aMaServerRequest ]! !

!MaTcpRequestServerLink methodsFor: 'accessing' stamp: 'cmm 10/8/2002 22:23'!
timeoutSeconds

	^socket timeoutSeconds! !

!MaTcpRequestServerLink methodsFor: 'accessing' stamp: 'cmm 10/8/2002 22:22'!
timeoutSeconds: anInteger

	socket timeoutSeconds: anInteger! !

MaObject subclass: #MaServerRequestLifecycle
	instanceVariableNames: 'socket socketQueueTime socketDequeueTime requestQueueTime requestDequeueTime timeToExtractRequestBytes requestByteArray beginSendingResponseTime doneSendingResponseTime responseByteArray remoteAddress remotePort'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma client server-Server API'!

!MaServerRequestLifecycle class methodsFor: 'create' stamp: 'cmm 5/1/2008 16:18'!
socket: aMaClientSocket 
	^ self new
		socket: aMaClientSocket ;
		yourself! !

!MaServerRequestLifecycle methodsFor: 'accessing' stamp: 'cmm 5/1/2008 19:19'!
beginSendingResponse
	beginSendingResponseTime := Time millisecondClockValue! !

!MaServerRequestLifecycle methodsFor: 'accessing' stamp: 'cmm 5/1/2008 19:18'!
beginSendingResponseTime
	^beginSendingResponseTime! !

!MaServerRequestLifecycle methodsFor: 'accessing' stamp: 'cmm 5/1/2008 19:18'!
beginSendingResponseTime: anInteger
	beginSendingResponseTime := anInteger! !

!MaServerRequestLifecycle methodsFor: 'copying' stamp: 'cmm 6/2/2008 13:15'!
copyForNewRequest
	^ self copy postCopyForNewRequest! !

!MaServerRequestLifecycle methodsFor: 'accessing' stamp: 'cmm 5/1/2008 19:26'!
doneSendingResponse
	doneSendingResponseTime := Time millisecondClockValue! !

!MaServerRequestLifecycle methodsFor: 'accessing' stamp: 'cmm 5/1/2008 16:22'!
millisecondsWaitingInSocketQueue
	^ Time 
		milliseconds: socketDequeueTime
		since: socketQueueTime! !

!MaServerRequestLifecycle methodsFor: 'initialize-release' stamp: 'cmm 5/1/2008 16:30'!
noteSocketDequeued
	socketDequeueTime := Time millisecondClockValue! !

!MaServerRequestLifecycle methodsFor: 'copying' stamp: 'cmm 6/2/2008 14:02'!
postCopyForNewRequest
	"The lifecycle stat is instantiated upon the initial socket connection, then cloned for each request to have its own stats."
	requestQueueTime := nil.
	requestDequeueTime := nil.
	timeToExtractRequestBytes := nil.
	requestByteArray := nil.
	beginSendingResponseTime := nil.
	doneSendingResponseTime := nil.
	responseByteArray := nil! !

!MaServerRequestLifecycle methodsFor: 'identification' stamp: 'cmm 5/6/2008 15:39'!
remoteAddress
	^ remoteAddress! !

!MaServerRequestLifecycle methodsFor: 'identification' stamp: 'cmm 5/6/2008 15:39'!
remotePort
	^ remotePort! !

!MaServerRequestLifecycle methodsFor: 'accessing' stamp: 'cmm 5/1/2008 16:37'!
requestByteArray
	^requestByteArray! !

!MaServerRequestLifecycle methodsFor: 'accessing' stamp: 'cmm 5/1/2008 16:37'!
requestByteArray: anObject
	requestByteArray := anObject! !

!MaServerRequestLifecycle methodsFor: 'initialize-release' stamp: 'cmm 5/1/2008 16:24'!
requestDequeueTime: anInteger
	requestDequeueTime := anInteger! !

!MaServerRequestLifecycle methodsFor: 'initialize-release' stamp: 'cmm 5/1/2008 16:24'!
requestQueueTime: anInteger
	requestQueueTime := anInteger! !

!MaServerRequestLifecycle methodsFor: 'accessing' stamp: 'cmm 5/1/2008 19:30'!
responseByteArray
	^responseByteArray! !

!MaServerRequestLifecycle methodsFor: 'accessing' stamp: 'cmm 5/1/2008 19:30'!
responseByteArray: aByteArray 
	responseByteArray := aByteArray! !

!MaServerRequestLifecycle methodsFor: 'identification' stamp: 'cmm 5/6/2008 15:39'!
sessionId
	^sessionId! !

!MaServerRequestLifecycle methodsFor: 'initialize-release' stamp: 'cmm 5/6/2008 15:39'!
sessionId: anObject
	sessionId := anObject! !

!MaServerRequestLifecycle methodsFor: 'accessing' stamp: 'cmm 5/1/2008 16:18'!
socket
	^socket! !

!MaServerRequestLifecycle methodsFor: 'initialize-release' stamp: 'cmm 5/6/2008 14:07'!
socket: aMaSocket 
	socket := aMaSocket.
	socketQueueTime := Time millisecondClockValue.
	remoteAddress := socket remoteAddress.
	remotePort := socket remotePort! !

!MaServerRequestLifecycle methodsFor: 'accessing' stamp: 'cmm 5/1/2008 16:41'!
timeToExtractRequestBytes
	^timeToExtractRequestBytes! !

!MaServerRequestLifecycle methodsFor: 'accessing' stamp: 'cmm 5/1/2008 16:42'!
timeToExtractRequestBytes: anInteger
	"Specified in anInteger milliseconds."
	timeToExtractRequestBytes := anInteger! !

!MaServerRequestLifecycle methodsFor: 'accessing' stamp: 'cmm 5/1/2008 19:27'!
timeToSendResponse
	^ Time 
		milliseconds: doneSendingResponseTime
		since: beginSendingResponseTime! !

!MaServerRequestLifecycle methodsFor: 'accessing' stamp: 'cmm 5/1/2008 17:58'!
timeWaitingInRequestQueue
	^ Time 
		milliseconds: requestDequeueTime
		since: requestQueueTime! !

!MaServerRequestLifecycle methodsFor: 'accessing' stamp: 'cmm 5/1/2008 17:42'!
timeWaitingInSocketQueue
	^ Time 
		milliseconds: socketDequeueTime
		since: socketQueueTime! !

MaObject subclass: #MaServerSocket
	instanceVariableNames: 'socketQueue requestQueue processRequestProcess socket timeoutSeconds doBlock port running server socketQueueMonitorProcess clientConnections clientConnectionsGuard'
	classVariableNames: 'ActiveSockets'
	poolDictionaries: ''
	category: 'Ma client server-Server API'!

!MaServerSocket commentStamp: 'cmm 6/6/2008 10:19' prior: 0!
I am the class which manages the underlying Squeak socket that listens for connecting client sockets.  I receive and process requests in the order received in the background and send responses off in their own background process as well.!

!MaServerSocket class methodsFor: 'initializing' stamp: 'cmm 4/30/2003 00:49'!
initialize

	ActiveSockets _ Set new.
	Smalltalk
		addToStartUpList: self;
		addToShutDownList: self! !

!MaServerSocket class methodsFor: 'initializing' stamp: 'cmm 12/27/2004 23:16'!
shutDown: aboutToQuit

	super shutDown: aboutToQuit.
	ActiveSockets do: [ :each | each primShutdown ]! !

!MaServerSocket class methodsFor: 'initializing' stamp: 'cmm 2/10/2005 09:42'!
startUp: amResuming

	super startUp: amResuming.
	ActiveSockets do: 
		[ :each | 
		"If another startUp method creates and starts one of me listening, then I don't want to tell it to listen again"
		each isRunning ifFalse: [ each listen ] ]! !

!MaServerSocket methodsFor: 'private' stamp: 'cmm 5/1/2008 18:05'!
dequeueByteArrayRequest
	| requestLifecycle |
	requestLifecycle := requestQueue next.
	requestLifecycle requestDequeueTime: Time millisecondClockValue.
	server noteRequestDequeued: requestLifecycle.
	^ requestLifecycle! !

!MaServerSocket methodsFor: 'private' stamp: 'cmm 6/8/2008 17:18'!
extractByteArrayRequestsFrom: aMaRequestLifecycle 
	"A socket connection has just been established.  Kickoff a new iterating process for the purpose of listening for requests bytes from a single client.  receiveOneByteArrayRequestFrom: waits a few seconds on the port, if data comes in it will extract the entire request and return it.  Otherwise it returns nil and the loop can check for shutdown or expiry."
	clientConnectionsGuard critical: 
		[ 1 
			to: clientConnections size - self maxConnections
			do: [ : n | (clientConnections at: n) shouldContinue: false ].
		clientConnections add: (MaClientConnection 
				serverSocket: self
				lifecycle: aMaRequestLifecycle) monitorForByteArrayRequests.
		server noteNumberOfConnections: clientConnections size ]! !

!MaServerSocket methodsFor: 'private' stamp: 'cmm 9/24/2002 23:01'!
headerSize
	"4-byte size header."

	^4! !

!MaServerSocket methodsFor: 'printing' stamp: 'cmm 10/2/2002 23:23'!
hostAndPortStringFor: aSocket

	^'host: ', aSocket remoteAddress maAsIpString, ' port: ', aSocket port printString! !

!MaServerSocket methodsFor: 'initialize-release' stamp: 'cmm 6/8/2008 11:14'!
initialize
	super initialize.
	running := false.
	self initializeQueues.
	timeoutSeconds := 30.
	self server: MaServer new.	"use the do-nothing server by default"
	doBlock := [ : byteArray | server processRequestByteArray: byteArray ].
	clientConnections := OrderedCollection new.
	clientConnectionsGuard := Monitor new! !

!MaServerSocket methodsFor: 'private' stamp: 'cmm 4/30/2003 00:30'!
initializeQueues

	socketQueue := SharedQueue new.
	requestQueue := SharedQueue new! !

!MaServerSocket methodsFor: 'testing' stamp: 'cmm 2/10/2005 09:42'!
isRunning

	^ running! !

!MaServerSocket methodsFor: 'private' stamp: 'cmm 5/30/2008 16:08'!
listen
	"By using this method, I will call my servers #processRequestByteArray:.  If you don't want to have a MaServer class, use listenOn:answer:."

	running ifTrue: [ MaClientServerUserError signal: 'socket already listening.' ].
	running _ true.
	ActiveSockets add: self.
	server note: 'initializing network drivers'.
	server note: 'creating server socket'.
	socket _ Socket newTCP.
	socket
		listenOn: port
		backlogSize: 50.
	socket isValid ifFalse: [ MaNetworkError signal: 'TCP/IP accept() not supported.' ].
	"Pop first socket off, which cannot accept data."
	socket 
		waitForAcceptFor: 5
		ifTimedOut: [ nil ].
	server note: 'now listening on ', port printString.
	self
		startAcceptingRequests;
		startExtractingRequests;
		startProcessingRequests! !

!MaServerSocket methodsFor: 'start/stop' stamp: 'cmm 4/30/2003 00:24'!
listenOn: portNumber
	"By using this method, I will call my servers #processRequestByteArray:.  If you don't want to have a MaServer class, use listenOn:answer:."

	port _ portNumber.
	self listen! !

!MaServerSocket methodsFor: 'start/stop' stamp: 'cmm 8/20/2005 21:36'!
listenOn: portNumber answer: oneArgBlock
	"Listen on portNumber, for each ByteArray request received answer the value of oneArgBlock (which should answer a ByteArray)."

	doBlock _ oneArgBlock.
	self listenOn: portNumber! !

!MaServerSocket methodsFor: 'private' stamp: 'cmm 6/8/2008 16:51'!
maxConnections
	^ 50! !

!MaServerSocket methodsFor: 'private' stamp: 'cmm 6/5/2008 23:21'!
primShutdown
	running := false.
	server note: 'new sockets no longer accepted.'.
	socketQueue isEmpty ifFalse: [ server note: 'waiting for queued sockets..' ].
	[ socketQueue isEmpty ] whileFalse: [ (Delay forSeconds: 1) wait ].
	socketQueueMonitorProcess terminate.
	requestQueue isEmpty ifFalse: [ server note: 'waiting for pending requests to process..' ].
	[ requestQueue isEmpty ] whileFalse: [ (Delay forSeconds: 1) wait ].
	processRequestProcess terminate.
	server note: 'deallocating resources.'.
	socket ifNotNil: [ socket closeAndDestroy ].
	self initializeQueues.
	socket := nil! !

!MaServerSocket methodsFor: 'private' stamp: 'cmm 5/1/2008 19:07'!
processRequest: aMaRequestLifecycle 
	| responseByteArray |
	responseByteArray := doBlock value: aMaRequestLifecycle requestByteArray.
	self 
		sendData: responseByteArray
		using: aMaRequestLifecycle! !

!MaServerSocket methodsFor: 'private' stamp: 'cmm 6/2/2008 22:14'!
queueRequest: aMaRequestLifecycle 
	running ifTrue: 
		[ aMaRequestLifecycle requestQueueTime: Time millisecondClockValue.
		"There is a slim possibility that, since entering this block, before the #nextPut, a #shutdown could have been issued, beginning the #primShutdown process.  I doubt it could happen, but just for extra safefy, one last check if we're still up.  In #primShutdown we set my 'socket' to nil.."
		socket ifNotNil: 
			[ requestQueue nextPut: aMaRequestLifecycle.
			server noteRequestQueued: aMaRequestLifecycle ] ]! !

!MaServerSocket methodsFor: 'private' stamp: 'cmm 6/2/2008 20:19'!
queueSocket: aSocket 
	socketQueue nextPut: (MaServerRequestLifecycle socket: aSocket).
	server noteSocketQueued: aSocket! !

!MaServerSocket methodsFor: 'private' stamp: 'cmm 6/8/2008 11:34'!
removeConnection: aMaClientConnection 
	clientConnectionsGuard critical: 
		[ clientConnections 
			remove: aMaClientConnection
			ifAbsent: 
				[ "not worth blowing up"
				 ].
		server noteNumberOfConnections: clientConnections size ]! !

!MaServerSocket methodsFor: 'private' stamp: 'cmm 6/8/2008 21:59'!
sendData: aByteArray using: aMaRequestLifecycle 
	
	[ | byteArrayToSend destinationSocket compressedByteArray |
	destinationSocket := aMaRequestLifecycle socket.
	destinationSocket isConnected 
		ifTrue: 
			[ server noteTimeToLoadBuffers: (Time millisecondsToRun: 
					[ compressedByteArray := aByteArray lzCompressed.
					byteArrayToSend := ByteArray new: compressedByteArray size + self headerSize.
					byteArrayToSend 
						maUint: 32
						at: 0
						put: byteArrayToSend size.
					byteArrayToSend 
						replaceFrom: self headerSize + 1
						to: byteArrayToSend size
						with: compressedByteArray
						startingAt: 1 ]).
			aMaRequestLifecycle responseByteArray: byteArrayToSend.
			aMaRequestLifecycle beginSendingResponse.
			server 
				noteSendingResponse: compressedByteArray
				to: destinationSocket.
			
			[ destinationSocket
				sendData: byteArrayToSend ;
				waitForSendDoneFor: timeoutSeconds ] 
				on: ConnectionTimedOut
				do: [ : exc | destinationSocket closeAndDestroy ].
			(destinationSocket isConnected and: [ destinationSocket sendDone ]) 
				ifTrue: 
					[ aMaRequestLifecycle doneSendingResponse.
					server noteResponseSent: aMaRequestLifecycle ]
				ifFalse: [ server noteClientTimedOut: aMaRequestLifecycle ] ]
		ifFalse: [ server noteClientTimedOut: aMaRequestLifecycle ] ] fork! !

!MaServerSocket methodsFor: 'private' stamp: 'cmm 6/8/2008 10:26'!
server
	^ server! !

!MaServerSocket methodsFor: 'server' stamp: 'cmm 12/30/2002 22:08'!
server: aMaServer
	"It is not required for the user to set this, a default one will be created."

	server _ aMaServer! !

!MaServerSocket methodsFor: 'start/stop' stamp: 'cmm 4/30/2003 00:46'!
shutdown

	ActiveSockets remove: self ifAbsent: [ "do nothing" ].
	self primShutdown! !

!MaServerSocket methodsFor: 'private' stamp: 'cmm 5/4/2008 23:54'!
startAcceptingRequests
	| clientSocket |
	
	[ [ running ] whileTrue: 
		[ clientSocket := socket 
			waitForAcceptFor: 5
			ifTimedOut: [ nil ].
		(clientSocket notNil and: [ clientSocket isConnected ]) ifTrue: [ self queueSocket: clientSocket ] ].
	server noteNoLongerAcceptingRequests ] newProcess
		name: 'Socket acceptor' ;
		priority: Processor userInterruptPriority ;
		resume.
	server note: 'Ready to queue sockets'! !

!MaServerSocket methodsFor: 'private' stamp: 'cmm 6/5/2008 23:21'!
startExtractingRequests
	socketQueueMonitorProcess := 
	[ [ running or: [ socketQueue notEmpty ] ] whileTrue: 
		[ | requestLifecycle |
		requestLifecycle := socketQueue next.
		requestLifecycle noteSocketDequeued.
		server noteSocketDequeued: requestLifecycle.
		self extractByteArrayRequestsFrom: requestLifecycle.
		requestLifecycle := nil ].
	server noteNoLongerExtractingRequests ] newProcess
		name: 'socketQueue Monitor' ;
		priority: Processor userInterruptPriority ;
		resume.
	server note: 'ready to extract requests'! !

!MaServerSocket methodsFor: 'private' stamp: 'cmm 8/27/2004 12:54'!
startProcessingRequests

	processRequestProcess _ 

		[

			[ running or: [ requestQueue notEmpty ] ]
				whileTrue: [ self processRequest: self dequeueByteArrayRequest ].
			server noteNoLongerProcessingRequests.

			] newProcess name: 'Request processor'; resume.
	server note: 'ready to process requests'! !

!MaServerSocket methodsFor: 'accessing' stamp: 'cmm 9/29/2002 21:33'!
timeoutSeconds

	^timeoutSeconds! !

!MaServerSocket methodsFor: 'accessing' stamp: 'cmm 9/29/2002 21:32'!
timeoutSeconds: anInteger

	timeoutSeconds _ anInteger! !

MaSerializationStat subclass: #MaClientStat
	instanceVariableNames: 'timeToConnectSocketHistory timeToLoadBuffersHistory timeToSendRequestsHistory timeWaitingForResponsesHistory timeReceivingResponsesHistory timeToSubmitByRequestClassHistorys'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma client server-Client API'!

!MaClientStat methodsFor: 'enumerate' stamp: 'cmm 5/9/2008 17:46'!
historysDo: aBlock 
	super historysDo: aBlock.
	{ 
		timeToConnectSocketHistory.
		timeToLoadBuffersHistory.
		timeToSendRequestsHistory.
		timeWaitingForResponsesHistory.
		timeReceivingResponsesHistory
	 } do: aBlock.
	timeToSubmitByRequestClassHistorys do: aBlock! !

!MaClientStat methodsFor: 'initialize-release' stamp: 'cmm 5/9/2008 17:45'!
initialize
	super initialize.
	timeToSubmitByRequestClassHistorys := Dictionary new.
	timeToConnectSocketHistory := MaStatHistory name: 'Time to Connect Sockets to Server'.
	timeToLoadBuffersHistory := MaStatHistory name: 'Time to Load Buffers'.
	timeToSendRequestsHistory := MaStatHistory name: 'Time to Send Requests'.
	timeWaitingForResponsesHistory := MaStatHistory name: 'Time for Initial Response'.
	timeReceivingResponsesHistory := MaStatHistory name: 'Time to Download Full Response'! !

!MaClientStat methodsFor: 'historys' stamp: 'cmm 5/9/2008 17:32'!
timeReceivingResponsesHistory
	"The time from when the initial response is received from the server to when all bytes of the response are downloaded.  By excluding the time waiting for the *initial* response, this stat attempts to measure the speed of downloading responses over the network."
	^timeReceivingResponsesHistory! !

!MaClientStat methodsFor: 'historys' stamp: 'cmm 5/9/2008 17:33'!
timeToConnectSocketHistory
	"Time to initiate a socket-connections, performed for for each request of the server."
	^timeToConnectSocketHistory! !

!MaClientStat methodsFor: 'historys' stamp: 'cmm 5/9/2008 17:33'!
timeToLoadBuffersHistory
	"For now Ma client server makes lots of copies of ByteArrays.  Until this is optimized, this stat tracks it to see how big of an impact it is."
	^timeToLoadBuffersHistory! !

!MaClientStat methodsFor: 'historys' stamp: 'cmm 5/9/2008 17:34'!
timeToSendRequestsHistory
	"The time needed just to send requests to the server."
	^timeToSendRequestsHistory! !

!MaClientStat methodsFor: 'historys' stamp: 'cmm 5/9/2008 17:45'!
timeToSubmitHistoryFor: requestClass 
	"The total time, by request class, including serialization, materialization, and all other house keeping.  A useful set of stats measuring the total reality for the client, by request type."
	^ timeToSubmitByRequestClassHistorys 
		at: requestClass
		ifAbsentPut: 
			[ MaStatHistory new
				name: 'Time waiting for ' , requestClass name , ' requests' ;
				yourself ]! !

!MaClientStat methodsFor: 'historys' stamp: 'cmm 5/9/2008 17:40'!
timeWaitingForResponsesHistory
	"When all bytes for a request have been pumped into the TCP/IP socket connected to the server, this stat immediately starts timing until we get the *first* initial response bytes back (which won't occur until the server has completely processed our request and is now sending back the response).  Therefore, this stat is a approximate measure of the time the server spends processing each request, including any time waiting in its request queue."
	^timeWaitingForResponsesHistory! !

MaSerializationStat subclass: #MaServerStat
	instanceVariableNames: 'startTime stopTime requestsQueuedHistory socketsQueuedHistory requestSizeHistory requestProcessingTimeHistory requestProcessingTimeHistorys responseSizeHistory extractTimeHistory timeWaitingInSocketQueueHistory timeWaitingInRequestQueueHistory timeToSendResponsesHistory clientTimeoutsHistory numberOfClientConnectionsHistory timeToLoadBuffersHistory'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma client server-Server API'!

!MaServerStat methodsFor: 'historys' stamp: 'cmm 5/7/2008 17:28'!
clientTimeoutsHistory
	"This history describes how often clients timed out.  When a request completes its lifecycle with no errors, a score of 0 is given.  When a network error occurs receiving the request or sending the response, a score of 1 is added.  Therefore, the sum is the total number of timeouts, #mean is how often they timed out."
	^ clientTimeoutsHistory! !

!MaServerStat methodsFor: 'recording' stamp: 'cmm 5/7/2008 17:29'!
decrementRequestsQueued
	| stat |
	stat := self requestsQueuedStat.
	stat last: stat last - 1! !

!MaServerStat methodsFor: 'recording' stamp: 'cmm 5/7/2008 17:29'!
decrementSocketsQueued
	| stat |
	stat := self socketsQueuedStat.
	stat last: stat last - 1! !

!MaServerStat methodsFor: 'historys' stamp: 'cmm 5/7/2008 17:29'!
extractTimeHistory
	"This history describes the amount of time, in milliseconds, elapsed from the initial client socket connection to the point in which the entire request is uploaded to my server."
	^ extractTimeHistory! !

!MaServerStat methodsFor: 'private' stamp: 'cmm 6/8/2008 21:35'!
historysDo: aBlock 
	super historysDo: aBlock.
	{ 
		socketsQueuedHistory.
		timeWaitingInSocketQueueHistory.
		extractTimeHistory.
		requestsQueuedHistory.
		timeWaitingInRequestQueueHistory.
		requestSizeHistory.
		requestProcessingTimeHistory.
		responseSizeHistory.
		timeToSendResponsesHistory.
		clientTimeoutsHistory.
		numberOfClientConnectionsHistory.
		timeToLoadBuffersHistory
	 } do: aBlock.
	requestProcessingTimeHistorys do: aBlock! !

!MaServerStat methodsFor: 'recording' stamp: 'cmm 5/7/2008 17:45'!
incrementRequestsQueued
	| history |
	history := self requestsQueuedHistory.
	history addValue: 1 + history last! !

!MaServerStat methodsFor: 'recording' stamp: 'cmm 5/7/2008 17:45'!
incrementSocketsQueued
	| history |
	history := self socketsQueuedHistory.
	history addValue: history last + 1! !

!MaServerStat methodsFor: 'initialize-release' stamp: 'cmm 6/8/2008 21:35'!
initialize
	super initialize.
	socketsQueuedHistory := MaStatHistory name: 'Sockets Queued'.
	requestsQueuedHistory := MaStatHistory name: 'Requests Queued'.
	requestSizeHistory := MaStatHistory name: 'Request Sizes'.
	requestProcessingTimeHistory := MaStatHistory name: 'Request Processing Times'.
	requestProcessingTimeHistorys := Dictionary new.
	responseSizeHistory := MaStatHistory name: 'Response Sizes'.
	extractTimeHistory := MaStatHistory name: 'Request Upload Times'.
	timeWaitingInSocketQueueHistory := MaStatHistory name: 'Time Waiting in Socket Queue'.
	timeWaitingInRequestQueueHistory := MaStatHistory name: 'Time Waiting in Request Queue'.
	timeToSendResponsesHistory := MaStatHistory name: 'Response Upload Times'.
	clientTimeoutsHistory := MaStatHistory name: 'Failed Lifecycles'.
	numberOfClientConnectionsHistory := MaStatHistory name: 'Number of Connected Client Sockets'.
	timeToLoadBuffersHistory := MaStatHistory name: 'Time to Compress and Uncompress'.
	startTime := DateAndTime now! !

!MaServerStat methodsFor: 'historys' stamp: 'cmm 6/8/2008 11:33'!
numberOfClientConnectionsHistory
	"The number of client sockets simultaneously connected to this server.  There is a Process running for each one that extracts the request bytes from them."
	^numberOfClientConnectionsHistory! !

!MaServerStat methodsFor: 'recording' stamp: 'cmm 6/5/2008 21:58'!
recordClientTimeout
	self clientTimeoutsHistory addValue: 1! !

!MaServerStat methodsFor: 'recording' stamp: 'cmm 6/5/2008 21:58'!
recordRequestSize: anInteger 
	self requestSizeHistory addValue: anInteger! !

!MaServerStat methodsFor: 'recording' stamp: 'cmm 6/5/2008 21:59'!
recordResponseSize: anInteger 
	self responseSizeHistory addValue: anInteger! !

!MaServerStat methodsFor: 'recording' stamp: 'cmm 6/5/2008 21:59'!
recordSuccessfulLifecycle
	self clientTimeoutsHistory addValue: 0! !

!MaServerStat methodsFor: 'recording' stamp: 'cmm 6/5/2008 21:59'!
recordTimeToExtractRequestBytes: milliseconds 
	self extractTimeHistory addValue: milliseconds! !

!MaServerStat methodsFor: 'recording' stamp: 'cmm 6/5/2008 21:59'!
recordTimeToSendResponse: anInteger 
	self timeToSendResponsesHistory addValue: anInteger! !

!MaServerStat methodsFor: 'recording' stamp: 'cmm 6/5/2008 21:59'!
recordTimeWaitingInRequestQueue: anInteger
	self timeWaitingInRequestQueueHistory addValue: anInteger! !

!MaServerStat methodsFor: 'recording' stamp: 'cmm 6/5/2008 21:59'!
recordTimeWaitingInSocketQueue: anInteger
	self timeWaitingInSocketQueueHistory addValue: anInteger! !

!MaServerStat methodsFor: 'historys' stamp: 'cmm 5/7/2008 17:33'!
requestProcessingTimeHistory
	"These statistics represent the time, in milliseconds, actually consumed by the Magma server to process client requests.  Although requests are processed one at a time, responses are sent in separate threads and request input downloads are also concurrent."
	^ requestProcessingTimeHistory! !

!MaServerStat methodsFor: 'historys' stamp: 'cmm 5/7/2008 17:34'!
requestProcessingTimeHistoryFor: requestObject 
	"Different requests take different amounts of time to process.  Track them by the class of request."
	^ requestProcessingTimeHistorys 
		at: requestObject class
		ifAbsentPut: 
			[ MaStatHistory new
				name: 'Time to process ' , requestObject class name ;
				yourself ]! !

!MaServerStat methodsFor: 'historys' stamp: 'cmm 5/7/2008 17:33'!
requestSizeHistory
	"The size of client requests, in bytes.  This would be related to extractTime, timeToMaterializeRequests and requestProcessingTimes."
	^ requestSizeHistory! !

!MaServerStat methodsFor: 'historys' stamp: 'cmm 5/7/2008 17:34'!
requestsQueuedHistory
	"The number of requests queued at each moment a new client request is queued.  How long does the request line get.."
	^ requestsQueuedHistory! !

!MaServerStat methodsFor: 'accessing' stamp: 'cmm 5/7/2008 17:32'!
requestsQueuedStat
	^ requestsQueuedHistory active! !

!MaServerStat methodsFor: 'historys' stamp: 'cmm 5/7/2008 17:34'!
responseSizeHistory
	"A statistic about how large the responses are.  As these numbers increase so, too, should the timeToSerializeResponses and timeToSendResponses increase."
	^ responseSizeHistory! !

!MaServerStat methodsFor: 'historys' stamp: 'cmm 5/7/2008 17:34'!
socketsQueuedHistory
	"The number of sockets in a queue waiting solely to have the bytes of the request extracted from the socket."
	^ socketsQueuedHistory! !

!MaServerStat methodsFor: 'accessing' stamp: 'cmm 5/7/2008 17:32'!
socketsQueuedStat
	^ socketsQueuedHistory active! !

!MaServerStat methodsFor: 'accessing' stamp: 'cmm 5/7/2008 17:32'!
startTime

	^ startTime! !

!MaServerStat methodsFor: 'accessing' stamp: 'cmm 5/7/2008 17:32'!
startTime: aMaCalendarTime

	startTime _ aMaCalendarTime! !

!MaServerStat methodsFor: 'accessing' stamp: 'cmm 5/7/2008 17:32'!
stopTime

	^ stopTime! !

!MaServerStat methodsFor: 'accessing' stamp: 'cmm 5/7/2008 17:32'!
stopTime: aMaCalendarTime

	stopTime _ aMaCalendarTime! !

!MaServerStat methodsFor: 'accessing' stamp: 'cmm 6/8/2008 21:33'!
timeToLoadBuffersHistory
	^timeToLoadBuffersHistory! !

!MaServerStat methodsFor: 'historys' stamp: 'cmm 5/7/2008 17:34'!
timeToSendResponsesHistory
	"The time required to send serialized responses out the TCP/IP port over the network.  Responses are sent in individual background threads."
	^ timeToSendResponsesHistory! !

!MaServerStat methodsFor: 'historys' stamp: 'cmm 5/7/2008 17:34'!
timeWaitingInRequestQueueHistory
	"The time, in milliseconds, from the point when the request is put into the request queue, all requests in front of it have been processed, and it is dequeued."
	^ timeWaitingInRequestQueueHistory! !

!MaServerStat methodsFor: 'historys' stamp: 'cmm 5/7/2008 17:34'!
timeWaitingInSocketQueueHistory
	"The duration, in milliseconds, from when the socket is queued until extraction of its ByteArray can begin.  All sockets in front of it get their ByteArray's extracted first."
	^ timeWaitingInSocketQueueHistory! !

!ByteArray methodsFor: '*ma client server' stamp: 'cmm 10/1/2002 19:42'!
maAsIpString

	^self size = 4
		ifTrue:
			[ (WriteStream on: (String new: 4))
				maPrint: (self at: 1);
				nextPut: $.;
				maPrint: (self at: 2);
				nextPut: $.;
				maPrint: (self at: 3);
				nextPut: $.;
				maPrint: (self at: 4);
				contents ]
		ifFalse: [ '(not an ip address)' ]! !

MaError subclass: #MaMalformedRequestError
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma client server-Exceptions'!

!MaMalformedRequestError commentStamp: 'cmm 2/1/2005 15:26' prior: 0!
I am signaled when a request is received from a client that could not be materialized.!

Object subclass: #MaClientConnection
	instanceVariableNames: 'serverSocket lifecycle extractProcess shouldContinue'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma client server-Server API'!

!MaClientConnection commentStamp: 'cmm 6/8/2008 10:09' prior: 0!
A MaClientConnection is created in the server image when a MaClientSocket connects to a MaServerSocket.  At that point, the client can send request ByteArrays which will be extracted by my extractProcess.  Up to 50 simultaneous connections will be allowed, after which the oldest will be disconnected (but not while they're in a request-lifecycle, of course, since they have to be able to send the response).

Instance Variables
	extractProcess:		<Process> - loops while the client #socket is connected, listening for request ByteArrays, extracting them and processing them when they become available.
	lifecycle:		<Object>
	processExpiryTimer:		<Object>
	serverSocket:		<Object>
	shouldContinue:		<Object>

extractProcess
	- xxxxx

lifecycle
	- xxxxx

processExpiryTimer
	- xxxxx

serverSocket
	- xxxxx

shouldContinue
	- xxxxx
!

!MaClientConnection class methodsFor: 'create' stamp: 'cmm 6/8/2008 10:46'!
serverSocket: aMaServerSocket lifecycle: aMaServerRequestLifecycle 
	^ self new
		serverSocket: aMaServerSocket ;
		lifecycle: aMaServerRequestLifecycle ;
		yourself! !

!MaClientConnection methodsFor: 'private' stamp: 'cmm 6/8/2008 10:51'!
ensureByteArray: aByteArray canHold: anInteger 
	| newAnswer |
	^ aByteArray size < anInteger 
		ifTrue: 
			[ newAnswer := ByteArray new: anInteger.
			newAnswer 
				replaceFrom: 1
				to: aByteArray size
				with: aByteArray
				startingAt: 1.
			newAnswer ]
		ifFalse: [ aByteArray ]! !

!MaClientConnection methodsFor: 'initialize-release' stamp: 'cmm 6/8/2008 10:44'!
initialize
	super initialize.
	shouldContinue := true! !

!MaClientConnection methodsFor: 'initialize-release' stamp: 'cmm 6/8/2008 10:46'!
lifecycle: aMaSeverRequestLifecycle
	lifecycle := aMaSeverRequestLifecycle! !

!MaClientConnection methodsFor: 'actions' stamp: 'cmm 6/8/2008 11:36'!
monitorForByteArrayRequests
	"A socket connection has just been established.  Kickoff a new iterating process for the purpose of listening for requests bytes from a single client (my lifecycles #socket).  receiveOneByteArrayRequestFrom: waits a few seconds on the port, if data comes in it will extract the entire request and return it.  Otherwise it returns nil and the loop can check for shutdown or expiry."
	extractProcess := 
	[ [ self shouldContinue and: [ lifecycle socket isConnected ] ] whileTrue: 
		[ | requestByteArray eachLifecycle |
		eachLifecycle := lifecycle copyForNewRequest.
		requestByteArray := [ self receiveOneByteArrayRequestFrom: eachLifecycle ] 
			on: Error
			do: 
				[ : exc | 
				self server 
					noteException: exc
					whileReceivingRequestFrom: eachLifecycle.
				self shouldContinue: false.
				nil ].
		requestByteArray ifNotNil: 
			[ eachLifecycle requestByteArray: requestByteArray.
			serverSocket queueRequest: eachLifecycle ] ].
	lifecycle socket closeAndDestroy.
	serverSocket removeConnection: self ] newProcess.
	extractProcess
		name: 'extracting requests from ' , lifecycle remoteAddress asString , ':' , lifecycle remotePort asString ;
		resume! !

!MaClientConnection methodsFor: 'private' stamp: 'cmm 6/8/2008 21:41'!
receiveOneByteArrayRequestFrom: aMaRequestLifecycle 
	"Wait up to 3 seconds for something from the client.  If data is received from aMaRequestLifecycle's socket be sure the entire data is received.  Otherwise answer nil."
	| bytesExpected bytesReceived answer firstDataReceived |
	bytesReceived := 0.
	bytesExpected := nil.
	answer := ByteArray new: 80.
	[ bytesExpected isNil or: [ bytesReceived < bytesExpected ] ] whileTrue: 
		[ aMaRequestLifecycle socket 
			waitForDataFor: 3
			ifClosed: 
				[ ConnectionClosed signal: 'Connection closed while waiting for data.' ]
			ifTimedOut: 
				[ "no data, just fall out"
				 ].
		aMaRequestLifecycle socket dataAvailable ifFalse: 
			[ "This logic means, if data starts to be received, each chunk must be received within 3 seconds or else it will timeout.  Is this ok?"
			bytesReceived = 0 
				ifTrue: [ ^ nil ]
				ifFalse: 
					[ MaNetworkError signal: 'data not available from ' , (serverSocket hostAndPortStringFor: aMaRequestLifecycle socket) ] ].
		firstDataReceived := Time millisecondClockValue.
		bytesReceived := bytesReceived + (aMaRequestLifecycle socket 
				receiveSomeDataInto: answer
				startingAt: bytesReceived + 1).
		bytesExpected ifNil: 
			[ bytesReceived >= serverSocket headerSize ifTrue: 
				[ bytesExpected := answer 
					maUint: 32
					at: 0.
				bytesExpected > MaObjectBuffer unreasonablePhysicalSize ifTrue: [ MaMalformedRequestError signal: 'unreasonable request size' ].
				answer := self 
					ensureByteArray: answer
					canHold: bytesExpected ] ] ].
	bytesReceived = bytesExpected ifFalse: 
		[ MaNetworkError signal: 'Unexpected client behavior, more bytes received than client stated sending.' ].
	self server noteTimeToExtractRequestBytes: (Time millisecondsSince: firstDataReceived).
	self server noteTimeToLoadBuffers: 
		(Time millisecondsToRun: 
			(answer := (answer 
				copyFrom: serverSocket headerSize + 1
				to: bytesExpected) lzUncompressed)).
	^ answer! !

!MaClientConnection methodsFor: 'private' stamp: 'cmm 6/8/2008 10:47'!
server
	^ serverSocket server! !

!MaClientConnection methodsFor: 'initialize-release' stamp: 'cmm 6/8/2008 10:45'!
serverSocket: aMaServerSocket
	serverSocket := aMaServerSocket! !

!MaClientConnection methodsFor: 'accessing' stamp: 'cmm 6/8/2008 10:43'!
shouldContinue
	^ shouldContinue and: [ serverSocket isRunning ]! !

!MaClientConnection methodsFor: 'accessing' stamp: 'cmm 6/8/2008 11:40'!
shouldContinue: aBoolean
	shouldContinue := aBoolean! !

PackageInfo subclass: #MaClientServerPackageInfo
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma client server-Info'!

!MaClientServerPackageInfo methodsFor: 'as yet unclassified' stamp: 'cmm 1/31/2003 00:20'!
additionalFiles

	^#('clientTest.ws' 'serverTest.ws')! !

!MaClientServerPackageInfo methodsFor: 'as yet unclassified' stamp: 'cmm 4/8/2004 23:13'!
changesText

	^ '
version 5:
	updated for 3.6 and 3.7.  This version does not work with 3.5.

version 4:
	running servers now survive image restarts.  The network is initialized by MaClientSocket upon starting the image, allowing sockets to immediately reconnect upon image start up.

version 3:
  performance improvement
  support for multiple clients in the same image connected to the same remote server

version 2:
  Should now support big and little endian machines simultaneously.

version 1:
  Initial public release.'! !

!MaClientServerPackageInfo methodsFor: 'accessing' stamp: 'cmm 4/8/2004 23:11'!
prerequisitePackages
	^ #( 'Ma object serialization' )! !

!MaClientServerPackageInfo methodsFor: 'as yet unclassified' stamp: 'cmm 1/30/2003 22:54'!
readmeText

	^'!!!!Ma client server

!!!!Introduction
I can help you write a client-server program very easily.

In the client-server model, a client sends a request to a server, the server processes the request and sends back a response.

This framework provides two layers interoperation.  You choose which layer you want to use and then write two programs.  A client program that will send requests to a remote server, which is running the other program to process those requests.

If you choose the lower layer, the "socket layer," you send ByteArray requests and receive ByteArray responses.  In the upper layer, the "object request layer" you send arbitrary request objects and get arbitrary objects as responses.  The "Ma object serialization" framework is used to convert the objects to ByteArrays which are sent and received for you using the socket layer.

!!!!Using the Socket Layer

!!Setting up the server
If you just need to send and receive ByteArray''s, you write your server program any way you want.  Just keep in mind that, somewhere, you''re going to want to take-in a ByteArray as input and you must answer a ByteArray as the response.

Then, to start the server:

=	myServerSocket _ MaServerSocket new
=		listenOn: 12345
=		answer:
=			[ :requestByteArray |
=			... give requestByteArray to your server program ... ]

MaServerSocket utilizes SharedQueue''s and background Processes to ensure that request processing is not held up by sending or receiving the ByteArray''s, which occurs entirely in the background.  The answer block is only ever evaluated for one client request at a time.

When you''re done, be sure to shutdown the socket so resources are properly deallocated:

=	myServerSocket shutdown

!!Getting answers in the client
Your client program creates a MaClientSocket by specifying the ip address as numbers in a ByteArray:

=	mySocket _ MaClientSocket hostAddress: #(192 168 1 1) port: 12345.

No connections are made just from creating the socket.  That happens when you actually send the request:

=	responseByteArray _ mySocket sendData: myRequestByteArray

There''s also a sendData:startingAt:count:waitForReplyIn: method for maximum efficiency.  If the reply ByteArray you supply wasn''t large enough, you''ll get a bigger one.

!!Socket Events
MaServerSocket knows about a "server", which defaults to a Server that does nothing (MaServer).  But you subclass MaServer and override some of its methods to learn when a socket is queued, a ByteArray request has been fully delivered, or sending a response, etc.  See MaServer for more information.

!!!!Using the Object Request Layer
If you choose to use the object request layer, you send "request" objects and get back arbitrary response objects.  So you write your client and server programs the same except you don''t concern yourself with parsing ByteArrays anymore.

!!Setting up the server
When you are ready to start your server program, instead of creating a MaServerSocket, you create a MaTcpRequestServer:

=	myServer _ MaTcpRequestServer protocol: { MyRequestClass1. MyRequestClass2.  MyResponseClass1.  MyResponseClass2.  OrderedCollection.  Dictionary. "etc." }.

What''s this?  Yes, you have to tell it all the classes that you''ll be sending and receiving, accounting for the full object-graph of your requests.  For example, if you have SubmitArticleRequest that has an Article object, you must include not only SubmitArticleRequest, Article and all the other kinds of objects that Article references, and what they reference, and so on.  You don''t however, need to ever specify String, ByteArray, Set, Symbol or Array.  Those are automatically part of every protocol.

You may also optionally specify a MaServerConsole to receive events.  Subclass MaServerConsole and override one of its notification methods and it''ll be called if you:

=	myServer _ console: myConsole

When you''re ready to start the server, you use processOn:do:

=	myServer
=		processOn: 12345
=		do:
=			[ :requestObject |
=			... pass requestObject to your server program, which answers the response ... ]

When you''re done, you need to shutdown to release operating system resources:

=	myServer shutdown

!!Setting up the client
Your client program creates a MaTcpRequestServerLink pointing to the server hostAddress and port is is listening:

=	myLinkToServer _ MaTcpRequestServerLink
=		hostAddress: #(192 168 1 1)
=		port: 12345

Before you can do anything, you have to connect.  This will ensure you have all the needed protocol in this client image (and throw an exception if you don''t).

=	myLinkToServer connect

To balance network fluctuations against your patience, you may wish to override the default timeout period from 30 seconds:

=	myLinkToServer timeoutSeconds: 30

!!Getting answers in the client
Now that you''re connected, your client program can get answers from the server by:

=	myLinkToServer submit: someRequestObject

Remember, someRequestObject will be a member of the protocol, as will the response you get back from calling submit:.

That''s it!!  When you''re done, don''t forget to disconnect so as you release network resources back to the operating system.

=	myLinkToServer disconnect
'! !

MaUserError subclass: #MaClientServerUserError
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma client server-Exceptions'!
MaClientServerPreferences initialize!
MaClientSocket initialize!
MaServerSocket initialize!
